using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX.DirectSound;
using System.Threading;
using fftwlib;
using System.Runtime.InteropServices;

namespace SoundReader
{
    public partial class Form1 : Form
    {
        private const int samplesToRead = 512;
        private const int samplingFrequency = 20050;

        // access this array only with lock(samples)
        private short[] samples = new short[samplesToRead];

        Thread thread;

        IntPtr pointerIn, pointerOut;
        float[] fourierIn, fourierOut;
        IntPtr fourierPlan;

        double lastAveraged = 0.0;
        double dcEstimate = 0.0;
        double previousDifference = 0;
        double[] fourierWindow = new double[samplesToRead];

        public Form1()
        {
            InitializeComponent();

            for (int i = 0; i < samplesToRead; i++)
            {
                // Hamming window
                fourierWindow[i] = 0.53836 - 0.46164 * Math.Cos(2.0 * Math.PI / (double)(samplesToRead-1) * i);
            }

            pointerIn = fftwf.malloc(samplesToRead * 4); // 4 bytes per float
            pointerOut = fftwf.malloc((samplesToRead + 2)* 4);
            fourierIn = new float[samplesToRead];
            fourierOut = new float[samplesToRead+2];
            fourierPlan = fftwf.dft_r2c_1d(samplesToRead, pointerIn, pointerOut, fftw_flags.Estimate);
           
            thread = new Thread(readAudio);
            thread.Start();
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            thread.Abort();

            fftwf.free(pointerIn);
            fftwf.free(pointerOut);
            fftwf.destroy_plan(fourierPlan);

            base.OnClosing(e);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            PointF[] points = new PointF[samplesToRead];
            PointF[] averagedPoints = new PointF[samplesToRead];
            double sum2 = 0.0;
            int numZC = 0;
            lock (samples)
            {
                for (int i = 0; i < samplesToRead; i++)
                {
                    // coarse shape of the signal:
                    // a zero line that glides up and down with the signal, so to speak
                    lastAveraged = 0.99 * lastAveraged + 0.01 * samples[i];
                    averagedPoints[i] = new PointF(i, 200.0f - 0.005f * (float)lastAveraged);
                    // an even slower average to determine the DC level
                    dcEstimate = 0.9999 * dcEstimate + 0.0001 * samples[i];

                    // count the crossings with the sliding zero line, not with the actual zero line
                    // to suppress low-frequency noise
                    double difference = samples[i] - lastAveraged;
                    if (difference * previousDifference <= 0.0) // sign change
                    {
                        numZC++;
                    }
                    previousDifference = difference;
                }

                for (int i = 0; i < samplesToRead; i++)
                {
                    double value = ((samples[i]-dcEstimate) / -(double)0x8000); // normalize to the range -1.0 ... +1.0
                    sum2 += value * value;
                }

                for (int i = 0; i < samplesToRead; i++)
                {
                    // apply the windowing function
                    fourierIn[i] = (float)(samples[i] * fourierWindow[i]);
                }
                Marshal.Copy(fourierIn, 0, pointerIn, samplesToRead);
                fftwf.execute(fourierPlan);
                Marshal.Copy(pointerOut, fourierOut, 0, samplesToRead + 1);

                for (int i = 0; i < samplesToRead; i++)
                {
                    points[i] = new PointF(i, 200.0f - 0.005f * (float)samples[i]);
                }
            }
            e.Graphics.DrawLines(Pens.Blue, averagedPoints);
            e.Graphics.DrawLines(Pens.Black, points);

            // TODO: We should compute spectra from sample blocks that overlap by half.
            PointF[] spectrum = new PointF[samplesToRead/2];
            for (int i = 0; i < samplesToRead/2; i++)
            {
                // power = real_part^2 + imaginary_part^2
                double real = fourierOut[2 * (i + 1)];
                double imag = fourierOut[2 * (i + 1)+1];
                double power = real*real + imag*imag;
                
                // level in decibels
                // +0.0001 to prevent Log10(0) = infinity
                // -80 to adjust vertical position
                double level = 10 * Math.Log10(power+0.0001)-80;
                
                // 2*i: stretch the display
                spectrum[i] = new PointF(2*i, 200.0f - 3.0f*(float)level);
            }
            e.Graphics.DrawLines(Pens.Red, spectrum);

            Text = (10.0*Math.Log10(sum2 / samplesToRead)).ToString("### dB") + " "
                    + (numZC * samplingFrequency / (double)samplesToRead).ToString("##### Zero Crossings /s");
        }

        private void readAudio()
        {
            using (Capture capture = new Capture())
            {
                CaptureBufferDescription desc = new CaptureBufferDescription();
                desc.BufferBytes = 4*2*samplesToRead; // room for four times the number of samples (each of two bytes)
                WaveFormat format = new WaveFormat();
                format.SamplesPerSecond = samplingFrequency;
                format.Channels = 1;
                format.BitsPerSample = 16; // data type: short
                format.FormatTag = WaveFormatTag.Pcm;
                format.BlockAlign = (short)(format.Channels * (format.BitsPerSample / 8));
                format.AverageBytesPerSecond = format.BlockAlign * format.SamplesPerSecond; 
                desc.Format = format;
                CaptureBuffer input = new CaptureBuffer(desc, capture);

                try
                {
                    input.Start(true);
                    int capturePosition;
                    int readPosition;
                    int lastReadPosition = 0;
                    while(true)
                    {
                        // TODO: Better use notification events instead of polling. I'm too lazy.
                        input.GetCurrentPosition(out capturePosition, out readPosition);
                        if (readPosition < lastReadPosition) // wrapped around?
                        {
                            readPosition += desc.BufferBytes;
                        }
                        if (readPosition > lastReadPosition + 2 * samplesToRead) // two bytes per sample
                        {
                            short[] data = (short[])input.Read(lastReadPosition, typeof(short), LockFlag.None, samplesToRead);
                            lock (samples)
                            {
                                Array.Copy(data, samples, samplesToRead);
                            }
                            lastReadPosition += 2 * samplesToRead; // two bytes per sample
                            lastReadPosition %= desc.BufferBytes; // wrap around
                            Invalidate();
                        }
                        else
                        {
                            Thread.Sleep(10);
                        }
                    }
                }
                finally
                {
                    input.Dispose();
                }
            }
        }
    }
}